/* 
Notes
	- Each ColorPicker instance represents a new view. All outside code should only need to interact with this object.
	- If you want to tweak the colors, make sure you spend some time looking at the hsl space. There's
		a few tweaks we i]put in here to make things look a little bit better.
Types
	Rgb - a 3 element array [r, g, b] with values fron 0 - 255.
	Hsl - a 3 element array [h, s, l] with values fron 0 - 1.

Example Usage
	var picker = new ColorPicker();
	$('body').append(picker.jq);
	picker.jq.css({
		left: 200,
		top: 200,
		width: 200,
		height: 200,
	});
	picker.layout();
	picker.setRgb([255, 0, 0]);
	picker.onHexClick = function(picker, newRgb) {
		
	};
	picker.onWheelUp = function(picker, newRgb) {
		
	};
	picker.onDrag = function(picker, newRgb) {
		
	};
	picker.onXClick = function(picker) {
		picker.hide();
	};
	

*/

ColorPicker = function() {
	//Properties
	this.jq = $('<div/>'); //Main wrapper
	this.hexs = [];
	this.wheel = $('<canvas/>');
	this.innerWheel = $('<div/>');
	this.spot = $('<div/>');
	this.innerSpot = $('<div/>');
	this.xView = $('<canvas/>');
	this._isDragging = false;
	this._spotDegrees = 0;
	this.onWheelUp = null; // function(ColorPicker p, Rgb newColor)
	this.onDrag = null; // function(ColorPicker p, Rgb newColor)
	this.onHexClick = null; // function(ColorPicker p, Rgb newColor)
	this.onXClick = null; // function(ColorPicker p)

	//Create Wheel
	this.wheel.css({
			position: 'absolute',
			cursor: 'pointer',
		});
	this.innerSpot.css({
		border: '2px solid white',
		borderRadius: '60%',
	});
	this.spot.css({
		position: 'absolute',
		cursor: 'pointer',
	});
	this.spot.append(this.innerSpot);
	this.innerWheel.css({
		position: 'absolute',
		backgroundColor: 'white',
	});
	this.jq.css({
		position: 'absolute',
		'-webkit-user-select': 'none',
	});
	this.jq.append(this.wheel, this.innerWheel, this.spot);

	//Create Hexs
	var i;
	for(i = 0; i < 8; i++) {
		var hex = new ColorPicker.Hex();
		hex.onClick = this._onHexClick.bind(this);
		hex.setBorder(0.05, 'black');
		if (i == 7) {
			hex.setTransparent(true);
		} else {
			hex.setTransparent(false);
		}
		this.jq.append(hex.jq);
		this.hexs.push(hex);
	}

	//Create X
	this.xView.css({
		position: 'absolute',
		right: 4,
		top: 4,
		width: 12,
		height: 12,
		cursor: 'pointer',
	});
	this.jq.append(this.xView);
	this.drawX();

	//Events
	this.xView.click(this._onXClick.bind(this));
	this.wheel.mousedown(this._onWheelMouseDown.bind(this));
	this.spot.mousedown(this._onWheelMouseDown.bind(this));
	this.innerWheel.mousedown(this.stopProp);
	this.jq.mousedown(this.stopProp)
		.mousemove(this._onMouseMove.bind(this))
		.mouseup(this._onMouseUp.bind(this))
		.mouseleave(this._onMouseUp.bind(this))
		.click(this.stopProp);
	this.setDegrees(0);
};


ColorPicker.prototype = {

	//Methods
	show: function() {
		this.jq.show();
	},

	hide: function() {
		this.jq.hide();
	},

	rgb: function() {
		return this.rgbForHex(this.hexs[3]);
	},

	setRgb: function(rgb) {
		var radians = this.rgbToRadians(rgb);
		this.setSpotDegrees(radians * 180 / Math.PI);
		this.setCenterHexHsl(ColorPicker.rgbToHsl(rgb));
	},

	setDegrees: function(degrees) {
		this.setSpotDegrees(degrees);
		this.setCenterHexHsl(this.radiansToHsl(degrees * Math.PI / 180));
	},

	//Layout
	layout: function(ops) {
		/* 
		ops { // optional - these are all percentages of the total width
			wheelWidth : Number -- 0 - 1
			hexWidth: Number -- 0 - 1
			wheelThickness: Number -- 0 - 1
			hexRadius: Number -- 0 - 1
		}
		*/
		//validate ops
		if(!ops) {
			ops = {};
		}
		if(ops.wheelWidth === undefined) {
			ops.wheelWidth = 0.8;
		}
		if(ops.hexWidth === undefined) {
			ops.hexWidth = 0.15;
		}
		if(ops.wheelThickness === undefined) {
			ops.wheelThickness = 0.1;
		}
		if(ops.hexRadius === undefined) {
			ops.hexRadius = 1.1 * ops.hexWidth;
		}
		var w = this.jq.width();
		this.layoutHexs(ops.hexWidth * w, ops.hexRadius * w);
		this.layoutColorWheel(ops.wheelWidth * w, ops.wheelThickness * w);
		this.layoutTransparentHex(ops.hexWidth * w, ops.wheelThickness * w, ops.wheelWidth * w, ops.wheelThickness * w);
	},

	layoutTransparentHex: function(hexWidth, rad, width, thick) {
		var w = this.jq.width();
		var cx = w / 2;
		var cy = w / 2;
		var angleStep = Math.PI / 3;
		var angle = (4 * Math.PI / 3) + (3 * angleStep);
		this.hexs[7].layout(w-rad-hexWidth+thick+thick * Math.cos(angle), w-rad-hexWidth+thick+5 * Math.sin(angle), hexWidth, true);
	},

	layoutHexs: function(hexWidth, rad) {
		var w = this.jq.width();
		var cx = w / 2;
		var cy = w / 2;
		var hex = this.hexs[3];
		this.hexs[3].layout(cx, cy, hexWidth);
		var angle = 4 * Math.PI / 3;
		var angleStep = Math.PI / 3;
		var order = [0, 1, 4, 6, 5, 2]; //clockwise order
		for(var i = 0; i < 6; i++) {
			this.hexs[order[i]].layout(cx + rad * Math.cos(angle), cy + rad * Math.sin(angle), hexWidth, false);
			angle += angleStep;
		}
	},

	layoutColorWheel: function(width, thick) {
		var w = this.jq.width();
		var top = 0.5 * w - width / 2;
		var left = top;
		var innerWidth = width - 2 * thick;
		this.wheel.css({
			left: left,
			top: top,
			width: width,
			height: width,
			borderRadius: width / 2,
		});
		this.wheel.attr('width', 2 * width)
			.attr('height', 2 * width);
		this.innerWheel.css({
			left: left + thick,
			top: top + thick,
			width: innerWidth,
			height: innerWidth,
			borderRadius: innerWidth / 2,
		});
		this.spot.css({
			left: left + thick + innerWidth,
			top: w / 2 - thick / 2,
			width: thick,
			height: thick,
			transformOrigin: -innerWidth / 2 + 'px ' + '50%',
		});
		this.innerSpot.css({
			width: thick - 4,
			height: thick - 4,
		});
		this.drawColorWheel();
	},

	//Color Wheel
	spotDegrees: function() {
		return this._spotDegrees;
	},

	setSpotDegrees: function(degrees) {
		degrees = degrees % 360;
		if(degrees < 0) {
			degrees += 360;
		}
		this._spotDegrees = degrees;
		this.spot.css({
			transform: 'rotate(' + degrees + 'deg)',
		});
	},

	drawColorWheel: function() {
        var canvas = this.wheel[0];
        var ctx = canvas.getContext('2d');
        var x = canvas.width / 2;
        var y = canvas.height / 2;
        var radius = canvas.width / 2;
        var step = 0.5;
        for (var a = 0; a <= 360; a += step) {
            var startRadians = (a - step) * Math.PI / 180;
            var endRadians = (a + step) * Math.PI / 180;
            var midRadians = a * Math.PI / 180;
            ctx.beginPath();
            ctx.moveTo(x, y);
            ctx.arc(x, y, radius, startRadians, endRadians, false);
            ctx.closePath();
            ctx.fillStyle = ColorPicker.hslToString(this.radiansToHsl(midRadians));
            ctx.fill();
        }
    },

    radiansToHsl: function(radians) {
    	var rads = this.normalizeRadians(radians + Math.PI / 2); 
        var v1 = 2/9 * Math.PI;
        var v2 = 14/9 * Math.PI;
        var v3 = 16/9 * Math.PI;
        if(rads < v1) {
            return [0, 1, 1 - 0.5 * (rads / v1)];
        } else if(rads > 1/3 * Math.PI && rads < 7/6 * Math.PI) {
        	//Lower the light here
        	var light = 0.35;
        	if(rads < 2/3 * Math.PI) {
        		light = 0.35 + 0.15 * (2/3 * Math.PI - rads) / (1/3 * Math.PI);
        	} else if(rads > Math.PI) {
        		light = 0.35 + 0.15 * (rads -  Math.PI) / (1/6 * Math.PI);
        	}
        	return [(rads - v1) / v3, 1, light];
        } else if(rads > v2 && rads <= v3) {
        	return [0.75, 1, 0.5 * (1 - (rads - v2) / v1)];
        } else if(rads > v3) {
        	return [0, 0, (rads - v3) / v1];
        } else {
        	return [(rads - v1) / v3, 1, 0.5];
        }
    },

    rgbToRadians: function(rgb) {
        var hsl = ColorPicker.rgbToHsl(rgb);
        var v1 = 2/9 * Math.PI;
        var v2 = 14/9 * Math.PI;
        var v3 = 16/9 * Math.PI;
        var radians = 0;
        if(hsl[1] === 0) {
        	//Black/White
        	radians = this.normalizeRadians(v3 + v1 * hsl[2])
        } else if(hsl[0] === 0) {
        	//White/Red
        	radians = 2 * (1 - hsl[2]) * v1;
        } else if(hsl[0] > 0.74 && hsl[0] < 0.76) {
        	//Purple
        	radians = v2 + 2 * (0.5 - hsl[2]) * v1;
        } else {
            radians = 16/9 * Math.PI * hsl[0] + 2/9 * Math.PI;
        }
        return radians - Math.PI / 2;
        // return hsl[0] * 2 * Math.PI - Math.PI / 2;
    },

    moveSpotToEventCoords: function(event) {
    	var off = this.jq.offset();
		var x = event.pageX - off.left - this.jq.height() / 2;
		var y = event.pageY - off.top - this.jq.width() / 2;
		var rads =  Math.atan2(y, x);
		this.setSpotDegrees(rads * 180 / Math.PI);
		this.setCenterHexHsl(this.radiansToHsl(rads));
    },

    //Hexs
    rgbForHex: function(hex) {
    	return ColorPicker.rgbFromString(hex.color());
    },

    setCenterHexRgb: function(rgb) {
    	this.setCenterHexHsl(ColorPicker.rgbToHsl(rgb))
    },

    setCenterHexHsl: function(hsl) {
    	//Get Shades
    	var shades = [];
        var dif = 0.5 - Math.abs(hsl[2] - 0.5); //distance to either 0 or 1, whichever is less.
        var step = dif / 4;
    	var light = hsl[2] + 3 * step;
    	var alpha = 100;
        for(var i = 0; i < 7; i++) {
        	shades.push([hsl[0], hsl[1], light, alpha]);
        	light -= step;
        }
        // shades[7] is transparent
        shades.push([0, 0, 100, 100]);

        //Set
        for(var i = 0; i < this.hexs.length; i++) {
        	var rgb = ColorPicker.hslToRgb(shades[i]);
        	this.hexs[i].setColor(ColorPicker.rgbToString(rgb));
        }
    },

    setHexBorder: function(width, color) {
    	for(var i = 0; i < this.hexs.length; i++) {
        	this.hexs[i].setBorder(width, color);
        }
    },

    //X
    drawX: function() {
    	var canvas = this.xView[0];
    	var w = 24;
    	var x1 = 4;
    	var x2 = w - x1;
    	// var thick = 2;
    	canvas.width = w;
    	canvas.height = w;
    	var ctx = canvas.getContext('2d');
    	ctx.lineWidth = 4;
    	ctx.lineCap = 'round';
    	ctx.strokeStyle = 'white';
    	ctx.beginPath();
	    ctx.moveTo(x1, x1);
	    ctx.lineTo(x2, x2);
	    ctx.moveTo(x1, x2);
	    ctx.lineTo(x2, x1);
	    ctx.stroke();
    },

    //Color Funcs
   	normalizeRadians: function(radians) {
   		radians = radians % (2 * Math.PI);
   		if(radians < 0) {
   			radians += 2 * Math.PI;
   		}
   		return radians;
   	},

	//Events
	_onHexClick: function(hex, event) {
		if(this.onHexClick) {
			this.onHexClick(this, this.rgbForHex(hex), hex.isTransparent());
		}
	},

	_onWheelMouseDown: function(event) {
		var off = this.jq.offset();
		var x = event.pageX - off.left - this.jq.height() / 2;
		var y = event.pageY - off.top - this.jq.width() / 2;
		var radius = Math.sqrt(x * x + y * y);
		if(radius < this.wheel.width() / 2) {
			this._isDragging = true;
			this.moveSpotToEventCoords(event);
			if(this.onDrag) {
				this.onDrag(this, this.rgb());
			}
			event.preventDefault();
		}
		
	},

	_onMouseMove: function(event) {
		if(this._isDragging) {
			this.moveSpotToEventCoords(event);
			if(this.onDrag) {
				this.onDrag(this, this.rgb());
			}
		}
		event.stopPropagation();
	},

	_onMouseUp: function(event) {
		if(this._isDragging && this.onWheelUp) {
			this.onWheelUp(this, this.rgb());
			
		}
		event.stopPropagation();
		this._isDragging = false;
	},

	_onXClick: function() {
		if(this.onXClick) {
			this.onXClick(this);
		}
		event.stopPropagation();
	},

	stopProp: function(event) {
		event.stopPropagation();
	},


};

/* Class representing one hex view. */
ColorPicker.Hex = function() {
	this.jq = $('<div/>');
	this.mid = $('<div/>');
	this.top = $('<div/>');
	this.bot = $('<div/>');
	this.canvas = $('<canvas/>');
	this.kids = [this.mid, this.top, this.bot];
	this.onClick = null; // function(ColorPicker.Hex hex)
	this._borderWidth = 0; //% of main width.
	this._borderColor = 'black';
	this._isTransparent = false;

	// checkerboard pattern for transparent hex
	this._checkerboardPattern = null;

	//Init
		for(var i = 0; i < this.kids.length; i++) {
		var k = this.kids[i];
		k.css({
			position: 'absolute',
			transformOrigin: 'top left',
			cursor: 'pointer',
		});
		k.click(this._onClick.bind(this));
		this.jq.append(k);
	}
    var trans = 'scaleY(0.5774) rotate(-45deg)';
		this.top.css({
		transform: trans,
		});
		this.bot.css({
		transform: trans,
		});
		this.jq.css({
		position: 'absolute',
		});
	this.canvas.css({
			position: 'absolute',
			width: '100%',
			height: '100%',
			pointerEvents: 'none',
		});
		// .attr('width', 200)
		// .attr('height', 200 * 2 / Math.sqrt(3));
	// this.drawBorder();
	this.jq.append(this.canvas);
};

ColorPicker.Hex.prototype = {

	color: function() {
		return this.mid.css('backgroundColor');
	},

	setColor: function(color) {
		for(var i = 0; i < this.kids.length; i++) {
			this.kids[i].css({ backgroundColor: color });
		}
	},

	isTransparent: function() {
		return this._isTransparent;
	},

	setTransparent: function(isTransparent) {
		this._isTransparent = isTransparent;
	},

	getCheckerboardPattern: function() {
		if (this._checkerboardPattern == null) {
			var pattern = document.createElement('canvas');
			pattern.width = 20;
			pattern.height = 20;
			var pctx = pattern.getContext('2d');

			var w=10;
	  		for(i=0; i < 2; i++) {
	    		for(j=0; j < 2; j++) {
	      			y = i * w;
	      			x = j * w;

	        		if(i % 2 != j % 2) {
						pctx.strokeStyle="#C0C0C0";
						pctx.fillStyle = "#C0C0C0";
	      			} else {
						pctx.strokeStyle="#FFFFFF";
						pctx.fillStyle = "#FFFFFF";
	      			}
	      			pctx.fillRect(x,y,w,w);
		        }
		    }

			pctx.fill();

			return pattern;
		} else {
			return this._checkerboardPattern;
		}
	},

	setBorder: function(width, color) {
		this._borderWidth = width;
		this._borderColor = color;
	},

	setWidth: function(width) {
		var height = width * 2 / Math.sqrt(3);
		var h1 = width / Math.sqrt(2);
		var h2 = width / Math.sqrt(3);
		var y1 = height / 2 - h2 / 2;
		this.jq.width(width)
			.height(height);
		this.mid.css({
			top: y1,
			width: width,
			height: h2,
		});
		this.top.css({
			top: y1,
			width: h1,
			height: h1,
		});
		this.bot.css({
			top: y1 + h2,
			width: h1,
			height: h1,
		});
	},

	setCenter: function(x, y) {
		this.jq.css({
			left: x - this.jq.width() / 2,
			top: y - this.jq.height() / 2,
		});
	},

	layout: function(cx, cy, width) {
		this.setWidth(width);
		this.setCenter(cx, cy);
		var canvasWidth = 2 * width;
		var canvasHeight = 2 * width * 2 / Math.sqrt(3);
		this.canvas.attr('width', canvasWidth)
			.attr('height', canvasHeight);
		this.drawBorder();
	},

	drawBorder: function() {
		var el = this.canvas[0];
		var width = el.width;
		var ctx = el.getContext('2d');
		ctx.lineWidth = this._borderWidth * width;
		ctx.strokeStyle = this._borderColor;
		ctx.lineJoin = 'round';
		ctx.lineCap = 'round';
		var radius = width * (0.5 - (this._borderWidth / 2)) * 2 / Math.sqrt(3);
		var radians = Math.PI / 6;
		var cx = width / 2;
		var cy = el.height / 2;
		var x = cx + radius * Math.cos(radians);
		var y = cy + radius * Math.sin(radians);

		ctx.beginPath();
		ctx.moveTo(x, y);
		for(var i = 0; i < 6; i++) {
			radians += Math.PI / 3;
			x = cx + radius * Math.cos(radians);
			y = cy + radius * Math.sin(radians);
			ctx.lineTo(x, y);
		}
		ctx.closePath();
		if (this._isTransparent) {
			var pattern = ctx.createPattern(this.getCheckerboardPattern(), "repeat");
			ctx.fillStyle = pattern;
			ctx.fill();
		}
		ctx.stroke();
	},

	_onClick: function(event) {
		if(this.onClick) {
			this.onClick(this);
		}
		event.stopPropagation();
	},
};

//Static Methods
ColorPicker.rgbToHsl =  function(rgb) {
    var r  = rgb[0] / 255;
    var g  = rgb[1] / 255;
    var b  = rgb[2] / 255;
    var max = Math.max(r, g, b);
    var min = Math.min(r, g, b);
    var h = 0;
    var s = 0;
    var l = (max + min) / 2;
    var d = max - min;
    if(d !== 0) {
        s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
        switch(max) {
            case r: 
            	h = (g - b) / d + (g < b ? 6 : 0); 
            	break;
            case g: 
            	h = (b - r) / d + 2; 
            	break;
            case b: 
            	h = (r - g) / d + 4; 
            	break;
        }
        h /= 6;
    }
    return [h, s, l];
};

ColorPicker.hslToRgb = function(hsl) {
	var r, g, b;
	var h = hsl[0];
	var s = hsl[1];
	var l = hsl[2];
    if(s === 0) {
        r = l; // achromatic
        g = l;
        b = l;
    } else {
        var q = l < 0.5 ? l * (1 + s) : l + s - l * s;
        var p = 2 * l - q;
        r = ColorPicker.hueToRgb(p, q, h + 1/3);
        g = ColorPicker.hueToRgb(p, q, h);
        b = ColorPicker.hueToRgb(p, q, h - 1/3);
    }
    return [Math.round(r * 255), Math.round(g * 255), Math.round(b * 255)];
};

ColorPicker.hueToRgb = function(p, q, t) {
	if(t < 0) {
		t += 1;
	} 
    if(t > 1) {
    	t -= 1;
    } 
    if(t < 1/6) {
    	return p + (q - p) * 6 * t;
    } 
    if(t < 1/2) {
    	return q;
    } 
    if(t < 2/3) {
    	return p + (q - p) * (2/3 - t) * 6;
    } 
	return p;
};

ColorPicker.rgbFromString = function(str) {
	if(str.substring(0, 4) !== 'rgb(') {
		return null;
	}
	var colors = str.substring(4, str.length - 1).split(',');
	var rgb = [];
	for(var i = 0; i < 3; i++) {
		rgb[i] = Number(colors[i]);
	}
	return rgb;
};

ColorPicker.rgbToString = function(rgb) {
	return 'rgb(' + rgb[0] + ',' + rgb[1] + ',' + rgb[2] + ')';
};

ColorPicker.hslToString = function(hsl) {
	return 'hsl(' + hsl[0] * 360 + ',' + hsl[1] * 100 + '%,' + hsl[2] * 100 + '%)';
};

ColorPicker.isDarkText = function(rgb) {
	/* Return true if text should be dark for the passed in background color. */
	var hsl = ColorPicker.rgbToHsl(rgb);
	var h = hsl[0];
	var threshold = 0.8;
	if(hsl[1] > 0.9 && h > 0.08 && h < 0.55) {
		//lower the threshold
		if(h < 0.12) {
			threshold = 0.56;
		} else if(h < 0.5) {
			threshold = 0.46;
		} else {
			threshold = 0.56 + 0.24 * (h - 0.5) / 0.05;
		}
	}
	return hsl[2] > threshold
};
