/* 
Main view class

Properties
	options: Option[]
	_count: int -- number of shown options
	_circleBorder: number -- % of main width
	_settingsBorders: number -- % of main width
	_pickerIndex: int -- option index corresponding to where  color picker is laid. -1 == no option
	_wedgeFontSize: number -- font size as % of circle width. If text is too big to fit at this size 
		it will get scaled down.
	picker: ColorPicker -- Only one instance of this view.
	settingsBtn: SettingsButton -- Button that is shown in top right
	onOptionInput: function(Event e, Option op) -- optional, called whenever user inputs text into otion input
	onColorBoxDown: function(Event e, Option op) -- optional, called on 
		mousedown.

Notes
	- Option objects are cached in mem when removed to preserve state (text & color)
		- This is why we store an extra _count variable to track how many options are
		actually shown.
	- Canvas is rendered at 2x so that the view looks better at different zoom levels
	- Angles are in radians where 0 is to the right of origin, and increase clockwise
*/

var SpinnerView = function(onOptionInput, onColorBoxDown) {
	this.options = [];
	this._count = 0;
	this._circleBorder = 0.008;
	this._settingsBorders = 0.01;
	this._pickerIndex = -1;
	this._wedgeFontSize = 0.07;
	this._circleMarg = 0.1;
	this._placeholder = ''; //Localized
	this.settingsBtn = new SettingsButton('res/gear.png');
	this.picker = new ColorPicker();
	this.onOptionInput = onOptionInput;
	this.onColorBoxDown = onColorBoxDown;
	this._checkerboardPattern = null;
	
	
	//Init
	this.picker.jq.addClass('picker');
	this.picker.setHexBorder(0.06, '#585858')
	$('#pickerWrap').append(this.picker.jq).hide();
	this.setSettingsShown(false);
	$('#back').css({
		opacity: 1,
	});
	$('#circle').prepend(this.settingsBtn.jq);
	$('#spinnerView').append($('#circle'), $('#pickerWrap'));
};

SpinnerView.prototype = {

	//Props
	initialize: function(settingsShown, opsData, count, pointerRadians) {
		this.setOpsData(opsData);
		this.setCount(count);
		this.setPointerRadians(pointerRadians);
		this.setSettingsShown(settingsShown);
		this.updatePointerColor();
	},

	firstAngle: function() {
		if(this._count === 3) {
			return 5 * Math.PI / 6;
		} else {
			return Math.PI / 2;
		}
	},

	circleMarg: function() {
		return this._circleMarg;
	},

	settingsShown: function() {
		return $('#back').is(':visible');
	},

	setSettingsShown: function(val) {
		if (val) {
			this.settingsBtn.back.css({
				backgroundColor: 'transparent',
			});
		} else {
			// hide the settings button background if there is a transparent wedge
			var hasTransparentWedge = this.hasTransparentWedge();
			this.settingsBtn.back.css({
				backgroundColor: hasTransparentWedge ? 'transparent' : '#03bede',
			});
		}
		if(val) {
			$('#back').show();
			$('#spinnerView').append(this.settingsBtn.jq);
		} else {
			$('#back').hide();
			$('#circle').prepend(this.settingsBtn.jq);
		}
		this.setPickerIndex(-1);
		this.settingsBtn.setShown(true);
	},

	pointerRadians: function() {
		return this._pointerRadians;
	},

	setPointerRadians: function(val) {
		val = val % (2 * Math.PI);
		this._pointerRadians = val;
		$('#pointerLine').css({
			transform: 'none',
		});
		var deg = 180 / Math.PI * val;
		$('#pointerLine').css({
			transform: 'rotate(' + deg + 'deg)',
		});
	},

	updatePointerColor: function() {
		var hasTransparentWedge = this.hasTransparentWedge();

		var els = ['#pointerLine', '#pointerCircle', '#pointerCircleBack'];
		var color = (hasTransparentWedge) ? 'black' : 'white';
		for (var i in els) {
			$(els[i]).css({
				"background-color": color,
			});
		}
	},

	setPlaceholder: function(val) {
		this._placeholder = val;
	},

	isSingleChar: function() {
		//Return true if every shown text is one or less
		for(var i = 0; i < this._count; i++) {
			if(this.options[i].text().length > 1) {
				return false;
			}
		}
		return true;
	},

	//Options
	count: function() {
		return this._count;
	},

	createOption: function() {
		var op = new Option();
		var data = {
			option: op,
		};
		op.input.on('input propertychange paste', data, this._onOptionInput.bind(this));
		op.input.attr('placeholder', this._placeholder);
		op.colorBox.on('mousedown', data, this._onColorBoxDown.bind(this));
		return op;
	},

	addOption: function() {
		if(this.options.length === this._count) {
			this.options.push(this.createOption());
		}
		this._count++;
		var op = this.options[this._count - 1];
		op.row.appendTo($('#optionsListBack'));
		this.layout();
		this.updateScrollBar();
	},

	opsData: function() {
		/* Includes cached options.
		Returns
			[{
				text: string,
				rgb: [number],
			}]
		*/
		var data = [];
		for(var i = 0; i < this.options.length; i++) {
			data.push({
				text: this.options[i].text(),
				rgb: this.options[i].rgb(),
				transparent: this.options[i].isTransparent,
			});
		}
		return data;
	},

	setOpsData: function(val) {
		/*
		
		*/
		//Add Necessaries
		var addN = val.length - this.options.length;
		for(var i = 0; i < addN; i++) {
			var op = this.createOption();
			this.options.push(op);
		}

		//Write
		for(var i = 0; i < val.length; i++) {
			var op = this.options[i];
			op.setText(val[i].text);
			this.setOptionRgb(op, val[i].rgb);
			op.setIsTransparent(val[i].transparent);
		}
	},

	removeOption: function() {
		var op = this.options[this._count - 1];
		op.row.detach();
		this._count--;
		if(this._pickerIndex > this._count) {
			this.setPickerIndex(-1);
		}
		this.layout();
		this.updateScrollBar();
	},

	setCount: function(n) {
		var oldN = this._count;
		var i;
		for(i = 0; i < oldN - n; i++) {
			this.removeOption();
		}
		for(i = 0; i < n - oldN; i++) {
			this.addOption();
		}
		this._count = n;
	},

	setOptionRgb: function(option, rgb) {
			this.setOptionRgb2(option, rgb, ColorPicker.isDarkText(rgb) ? '#6a6a6a' : 'white');
	},

	setOptionRgb2: function(option, rgb, textColor) {
		option.setRgb(rgb);
		option.textColor = textColor;
		this.drawWedges();
	},

	setOptionTransparency: function(option, isTransparent) {
		option.setIsTransparent(isTransparent);
	},

	//Picker
	pickerIndex: function() {
		return this._pickerIndex;
	},

	setPickerIndex: function(val) {
		var wrap = $('#pickerWrap');
		this._pickerIndex = val;
		if(val == -1) {
			wrap.hide();
		} else {
			wrap.show();
			this.picker.setRgb(this.options[val].rgb());
			this.layoutPicker();
		}
	},

	togglePicker: function(option) {
		/* Show picker if it is not shown for the passed in option, or hide it if it is. */
		var index = this.options.indexOf(option);
		if(index === this.pickerIndex()) {
			this.setPickerIndex(-1);
		} else {
			this.setPickerIndex(index);
		}
	},

	//Layout/Draw
	layout: function() {
		var width = $('#spinnerView').width();
		this.layoutCircle();
		if(this.settingsShown()) {
			this.settingsBtn.layoutInCorner(width);
		} else {
			this.settingsBtn.layoutBehindCircle($('#circle').width());
		}

		var h = $('#plusMinus').height();
		$('#plus, #minus').width(h).height(h);
		$('#segments').css({
			fontSize: 0.5 * h,
		});
		var border = this._settingsBorders * width;
		$('#plusMinus').css({
			borderWidth: border,
			borderRadius: border,
		});
		var lineHeight = 0.02 * $('#circle').width();
		$('#pointerLine').css({
			top: $('#circle').height() / 2 - lineHeight / 2,
			height: lineHeight,
		});

		this.layoutOptionsList();
		this.layoutPicker();
	},

	layoutCircle: function() {
		var cir = $('#circle');
		var mainWidth = $('#spinnerView').width();
		var w = 0;
		var x = 0;
		var marg = this.circleMarg();
		if(this.settingsShown()) {
			w = mainWidth * (1 - 2 * this._circleMarg);
			x = mainWidth * this._circleMarg;
		} else {
			w = mainWidth - 4;
			x = 2;
		}
		cir.css({
			width: w,
			height: w,
			left: x,
			top: x,
		});
		$('#canvas').attr('width', 2 * w)
			.attr('height', 2 * w);
		this.drawWedges();
	},	

	layoutOptionsList: function() {
		var list = $('#optionsList');
		var listHeight = list.height();
		var rowHeight = listHeight * 0.22;
		var gap = listHeight * 0.04;
		var y = 0;
		var fontSize = rowHeight * 0.4;
		var overflow = this._count > 4;
		var border = this._settingsBorders * $('#spinnerView').width();
		for(var i = 0; i < this._count; i++) {
			var op = this.options[i];
			op.row.css({
				top: y,
				right: overflow ? 3 : 0,
				height: rowHeight,
				borderWidth: border,
				borderRadius: border,
			});
			op.input.css({
				fontSize: fontSize,
			});
			op.layout();
			y += rowHeight + gap;
		}
		$('#optionsListBack').height(y - gap + 2); //Add 2 to adjust for rounding errors in layout

		//Overflow
		$('#optionsList').css({
			overflowY: overflow ? 'scroll' : 'hidden',
		});
	},

	layoutPicker: function() {
		var op = this.options[this._pickerIndex];
		if(op) {
			var off = op.colorBox.offset();
			var boxWidth = op.colorBox.width();
			var x = off.left + 0.5 * boxWidth;
			var y = off.top + op.colorBox.height() / 2;
			var right = $('#spinnerView').width() * 0.95;
			var width = right - x;
			$('#pickerWrap').css({
				left: x,
				top: y,
			});
			this.picker.jq.width(width).height(width);
			this.picker.layout({
				wheelWidth : 0.85,
				hexWidth: 0.15,
				wheelThickness: 0.12,
			});
		}
	},

	updatePickerPos: function() {
		var op = this.options[this._pickerIndex];
		if(op) {
			if(!op.isHalfShown()) {
 				this.setPickerIndex(-1);
			} else {
				this.layoutPicker();
			}
		}	
	},

	drawWedges: function() {
		var n = this._count;
		var ctx = $('#canvas')[0].getContext('2d');
		var w = $('#canvas')[0].width;
		var borderWidth = w * this._circleBorder;
		var radius = w / 2;
		var dif = 2 * Math.PI / n;

		var hasTransparentWedge = false;

		//Wedge Background
		var radians = this.firstAngle();
		for(var i = 0; i < n; i++) {
			var endRadians = radians + dif;
			if (this.options[i].isTransparent) {
				if (!this.settingsShown()) {
					ctx.setFillColor(0,0,0,0);
				} else {
					var pattern = ctx.createPattern(this.getCheckerboardPattern(), "repeat");
					ctx.fillStyle = pattern;
				}
				hasTransparentWedge = true;
			} else {
				ctx.fillStyle = rgb2Style(this.options[i].rgb());
			}
			ctx.beginPath();
			ctx.moveTo(radius, radius);
			ctx.arc(radius, radius, radius - borderWidth, radians, endRadians, false);
			ctx.fill();
			radians += dif;
		}

		//Text
		ctx.save();
		ctx.textBaseline = 'middle';
		ctx.textAlign = 'center';
		ctx.translate(w / 2, w / 2);
		if(n === 1) {
			var wMax = radius * 0.8;
			this.drawText(ctx, 0, 0, -radius / 2, 0, wMax);
		} else if(n === 2) {
			var wMax = radius * 0.8;
			this.drawText(ctx, 0, 0, -radius / 2, 0, wMax);
			this.drawText(ctx, 1, 0, radius / 2, 0, wMax);
		} else if(n === 3) {
			var wMax = radius * 0.8;
			var y1 = -radius / 3;
			var x1 = 0.9 * radius / 2;
			var wMax = radius * 0.8;
			this.drawText(ctx, 0, 0, -x1, y1, wMax);
			this.drawText(ctx, 1, 0, x1, y1, wMax);
			this.drawText(ctx, 2, 0, 0, radius / 2, wMax);
		} else if(n > 3) {
			var wMax = (0.9 - 0.03 * (n - 2)) * radius;
			var radians = this.firstAngle() + dif / 2;
			var x1 = 0.95 * radius - 0.5 * wMax;
			if(this.isSingleChar()) {
				//Dont rotate
				for(var i = 0; i < n; i++) {
					var x = x1 * Math.cos(radians);
					var y = x1 * Math.sin(radians);
					this.drawText(ctx, i, 0, x, y, wMax);
					radians += dif;
				}
			} else {
				for(var i = 0; i < n; i++) {
					this.drawText(ctx, i, radians, x1, 0, wMax);
					radians += dif;
				}
			}
		}
		ctx.restore();

		//Inner Border
		ctx.strokeStyle = (hasTransparentWedge) ? '#000000' : '#FFFFFF';
		ctx.lineWidth = borderWidth;
		radians = this.firstAngle();
		var r = radius - borderWidth;
		if (n > 1) {
			for(var i = 0; i < n; i++) {
				ctx.beginPath();
				ctx.moveTo(radius, radius);
				ctx.lineTo(radius + r * Math.cos(radians), radius + r * Math.sin(radians))
				ctx.stroke();
				radians += dif;
			}
		}

		//Outer Border
		ctx.beginPath();
		ctx.arc(radius, radius, 0.99 * radius - borderWidth / 2, 0, 2 * Math.PI, false);
		ctx.stroke();
	},

	drawText: function(ctx, index, radians, cx, cy, maxWidth) {
		// note that rotation is applied beofre text is translated to (cx, cy)
		var w = $('#canvas')[0].width;
		var text = this.options[index].input.val();
		var color = 'white';

		//Normalize Angle
		radians = radians % (2 * Math.PI);
		if(radians < 0) {
			radians += 2 * Math.PI;
		}
		if(radians > Math.PI / 2 + 0.01 && radians < 3 * Math.PI / 2 - 0.01) {
			radians = (radians + Math.PI) % (2 * Math.PI);
			cx = -cx;
		}

		ctx.save();
		ctx.fillStyle = this.options[index].textColor;
		ctx.rotate(radians);

		//Text Width
		var fontSize = Math.floor(this._wedgeFontSize * w);
		ctx.font = fontSize + 'px Arial';
		textWidth = ctx.measureText(text).width;
		if(textWidth > maxWidth) {
			fontSize = Math.floor(fontSize * maxWidth / textWidth);
			ctx.font = fontSize + 'px Arial';
		}

		//Draw
		ctx.fillText(text, cx, cy);
		ctx.restore();
	},

	updateScrollBar: function() {
		/* Hack to deal with scroll bar rendering issues on Win */
		var list = $('#optionsList');
		var sy = list.scrollTop();
		list.scrollTop(sy - 1);
		list.scrollTop(sy + 1);
		list.scrollTop(sy);
	},

	//Event Piping
	_onOptionInput: function(event) {
		if(this.onOptionInput) {
			this.onOptionInput(event, event.data.option);
		}
	},

	_onColorBoxDown: function(event) {
		if(this.onColorBoxDown) {
			var op = event.data.option;
			this.onColorBoxDown(event, op);
		}
	},

	//Other
	setPlusMinusValid: function(plusValid, minusValid) {
		$('#plus').css({
			opacity: plusValid ? 1 : 0.5,
		});
		$('#minus').css({
			opacity: minusValid ? 1 : 0.5,
		});
	},

	getCheckerboardPattern: function() {
		if (this._checkerboardPattern == null) {
			var pattern = document.createElement('canvas');
			pattern.width = 20;
			pattern.height = 20;
			var pctx = pattern.getContext('2d');

			var w=10;
	  		for(i=0; i < 2; i++) {
	    		for(j=0; j < 2; j++) {
	      			y = i * w;
	      			x = j * w;

	        		if(i % 2 != j % 2) {
						pctx.strokeStyle="#C0C0C0";
						pctx.fillStyle = "#C0C0C0";
	      			} else {
						pctx.strokeStyle="#FFFFFF";
						pctx.fillStyle = "#FFFFFF";
	      			}
	      			pctx.fillRect(x,y,w,w);
		        }
		    }

			pctx.fill();

			return pattern;
		} else {
			return this._checkerboardPattern;
		}
	},

	hasTransparentWedge: function() {
		var n = this._count;
		var hasTransparentWedge = false;
		for(var i = 0; i < n; i++) {
			if (this.options[i].isTransparent) {
				hasTransparentWedge = true;
				break;
			}
		}
		return hasTransparentWedge;
	},
};
